
/* Plan 9 module implementation */

static char plan9__doc__ [] =
"This module provides access to operating system functionality that\n\
is presented by Plan 9 from Bell Labs.  Refer to the corresponding manual\n\
entries for more information on calls.  Some calls are emulations of POSIX\n\
features; refer to a Unix manual for information on those.";

#include "Python.h"

#include "structseq.h" /* Added by Moroo. this should be put into Python.h */
/* copy from Python 2.4 Moroo */
#define STRUCT_STAT Dir
PyObject *PyStructSequence_New(PyTypeObject *type);
static PyTypeObject StatResultType;

static PyObject*
plan9_error(void)
{
	char buf[ERRMAX];

	rerrstr(buf, sizeof buf);
	PyErr_SetString(PyExc_OSError, buf);
	return nil;
}
 
/* pack a system stat C structure into the Python stat tuple  */
static PyObject*
dir2pystat(Dir *d)
{
	return Py_BuildValue("(s(Llb)(ci)lllLsss)",
		d->name, d->qid.path, d->qid.vers, d->qid.type,
		d->type, d->dev, d->mode, d->atime, d->mtime,
		d->length, d->uid, d->gid, d->muid);
}

static int
getvlong(PyObject *o, vlong *p)
{
	if(PyLong_Check(o))
		*p = PyLong_AsLongLong(o);
	else
		*p = (vlong)PyInt_AsLong(o);
	if(PyErr_Occurred())
		return 0;
	return 1;
}

static int
pystat2dir(PyObject *dt, Dir *d)
{
	PyObject *o, *o2;

	if(!PyArg_ParseTuple(dt, "s(Olb)(ci)lllOsss:(f)wstat",
		&d->name, &o, &d->qid.vers, &d->qid.type,
		&d->type, &d->dev, &d->mode, &d->atime, &d->mtime,
		&o2, &d->uid, &d->gid, &d->muid)
	|| !getvlong(o, (vlong*)&d->qid.path)
	|| !getvlong(o2, (vlong*)&d->length))
		return 0;
	return 1;
}

static char plan9_access__doc__[] =
"access(path, mode) -> 1 if granted, 0 otherwise\n\
Test for access to a file.";

static PyObject *
plan9_access(PyObject *self, PyObject *args)
{
	char *path;
	int mode;
	int res;

	if (!PyArg_ParseTuple(args, "si:access", &path, &mode))
		return NULL;
	Py_BEGIN_ALLOW_THREADS
	res = access(path, mode);
	Py_END_ALLOW_THREADS
	return(PyInt_FromLong(res == 0 ? 1L : 0L));
}

static char plan9_bind__doc__[] =
"bind(name, old, flag) -> int\n\
Change name space.";

static PyObject*
plan9_bind(PyObject *self, PyObject *args)
{
	char *name, *old;
	int flag, res;

	if (!PyArg_ParseTuple(args, "ssi:bind", &name, &old, &flag))
		return nil;
	Py_BEGIN_ALLOW_THREADS
	res = bind(name, old, flag);
	Py_END_ALLOW_THREADS
	if (res < 0)
		return plan9_error();
	return PyInt_FromLong(res);
}

static char plan9_chdir__doc__[] =
"chdir(path) -> None\n\
Change the current working directory to the specified path.";

static PyObject *
plan9_chdir(PyObject *self, PyObject *args)
{
	char *path;
	int res;

	if (!PyArg_ParseTuple(args, "s:chdir", &path))
		return nil;
	Py_BEGIN_ALLOW_THREADS
	res = chdir(path);
	Py_END_ALLOW_THREADS
	if (res < 0)
		return plan9_error();
	Py_INCREF(Py_None);
	return Py_None;
}

static char plan9_getwd__doc__[] =
"getwd() -> path\n\
Return a string representing the current working directory.";

static PyObject *
plan9_getwd(PyObject *self, PyObject *args)
{
	char buf[1024];
	char *res;
	if (!PyArg_ParseTuple(args, ":getcwd"))
		return NULL;
	Py_BEGIN_ALLOW_THREADS
	res = getwd(buf, sizeof buf);
	Py_END_ALLOW_THREADS
	if (res == nil)
		return plan9_error();
	return PyString_FromString(buf);
}


static char plan9_dirread__doc__[] =
"dirread(fd) -> list_of_dir_tuples\n\
Return a list containing tuple stat structures for some of the directory.\n\
\n\
	fd: fd of directory to list\n\
\n";

static PyObject *
plan9_dirread(PyObject *self, PyObject *args)
{
	int fd, i, n;
	Dir *d;
	PyObject *pd, *v;

	if (!PyArg_ParseTuple(args, "i:dirread", &fd))
		return nil;
	if((d = dirfstat(fd)) == nil)
		return plan9_error();
	if(!(d->mode&DMDIR)){
		free(d);
		PyErr_SetString(PyExc_OSError, "not a directory");
		return nil;
	}
	free(d);
	
	if ((pd = PyList_New(0)) == nil)
		return nil;

	n = dirread(fd, &d);
	if(n < 0){
		if(pd)
			Py_DECREF(pd);
		return plan9_error();
	}
	for(i=0; i<n; i++){
		v = dir2pystat(&d[i]);
		if(v == nil){
			Py_DECREF(pd);
			pd = nil;
			free(d);
			break;
		}
		if(PyList_Append(pd, v) != 0){
			Py_DECREF(v);
			Py_DECREF(pd);
			pd = nil;
			free(d);
			break;
		}
		Py_DECREF(v);
	}
	free(d);
	return pd;
}

static char plan9_dirreadall__doc__[] =
"dirreadall(path) -> list_of_dir_tuples\n\
Return a list containing tuple stat structures for the directory.\n\
\n\
	path: path of directory to list\n\
\n";

static PyObject *
plan9_dirreadall(PyObject *self, PyObject *args)
{
	int fd, i, n;
	char *name;
	Dir *d;
	PyObject *pd, *v;

	if (!PyArg_ParseTuple(args, "s:listdir", &name))
		return nil;
	if((fd = open(name, OREAD)) < 0)
		return plan9_error();
	if((d = dirfstat(fd)) == nil){
		close(fd);
		return plan9_error();
	}
	if(!(d->mode&DMDIR)){
		close(fd);
		free(d);
		PyErr_SetString(PyExc_OSError, "not a directory");
		return nil;
	}
	free(d);
	
	if ((pd = PyList_New(0)) == NULL){
		close(fd);
		return nil;
	}

	while((n = dirread(fd, &d)) > 0){
		for(i=0; i<n; i++){
			v = dir2pystat(&d[i]);
			if(v == nil){
				Py_DECREF(pd);
				pd = nil;
				free(d);
				goto Break2;
			}
			if(PyList_Append(pd, v) != 0){
				Py_DECREF(v);
				Py_DECREF(pd);
				pd = nil;
				free(d);
				goto Break2;
			}
			Py_DECREF(v);
		}
		free(d);
	}
Break2:
	close(fd);
	if(n < 0){
		if(pd)
			Py_DECREF(pd);
		return plan9_error();
	}
	return pd;
}


static char plan9_mount__doc__[] =
"mount(fd, afd, old, flag, aname) -> int\n\
Change name space.";

static PyObject*
plan9_mount(PyObject *self, PyObject *args)
{
	char *old, *aname;
	int fd, afd, flag, res;

	if (!PyArg_ParseTuple(args, "iisis:mount", &fd, &afd, &old, &flag, &aname))
		return nil;
	Py_BEGIN_ALLOW_THREADS
	res = mount(fd, afd, old, flag, aname);
	Py_END_ALLOW_THREADS
	if (res < 0)
		return plan9_error();
	return PyInt_FromLong(res);
}


static char plan9_stat__doc__[] =
"stat(path) -> dir-tuple\n\
Perform a stat system call on the given path.";

static PyObject *
plan9_stat(PyObject *self, PyObject *args)
{
	PyObject *ret;
	Dir *d;
	char *path;

	if(!PyArg_ParseTuple(args, "s:stat", &path))
		return nil;
	Py_BEGIN_ALLOW_THREADS
	d = dirstat(path);
	Py_END_ALLOW_THREADS
	if(d == nil)
		return plan9_error();
	ret = dir2pystat(d);
	free(d);
	return ret;
}

static char plan9_fstat__doc__[] =
"fstat(fd) -> dir-tuple\n\
Stat a fd.";

static PyObject*
plan9_fstat(PyObject *self, PyObject *args)
{
	PyObject *ret;
	Dir *d;
	int fd;

	if(!PyArg_ParseTuple(args, "i:fstat", &fd))
		return nil;
	Py_BEGIN_ALLOW_THREADS
	d = dirfstat(fd);
	Py_END_ALLOW_THREADS
	if(d == nil)
		return plan9_error();
	ret = dir2pystat(d);
	free(d);
	return ret;
}

static char plan9_wstat__doc__[] =
"wstat(path, (dir-tuples)) -> None\n\
Perform a wstat system call on the given path.";

static PyObject *
plan9_wstat(PyObject *self, PyObject *args)
{
	PyObject *dt;
	Dir d;
	char *path;
	int ret;

	if(!PyArg_ParseTuple(args, "sO:wstat", &path, &dt))
		return nil;
	if(!pystat2dir(dt, &d))
		return nil;
	Py_BEGIN_ALLOW_THREADS
	ret = dirwstat(path, &d);
	Py_END_ALLOW_THREADS
	if(ret < 0)
		return plan9_error();
	Py_INCREF(Py_None);
	return Py_None;
}

static char plan9_fwstat__doc__[] =
"fwstat(fd, (dir-tuples)) -> None\n\
Perform a fwstat system call on the given fd.";

static PyObject *
plan9_fwstat(PyObject *self, PyObject *args)
{
	PyObject *dt;
	Dir d;
	int fd;
	int ret;

	if(!PyArg_ParseTuple(args, "iO:fwstat", &fd, &dt))
		return nil;
	if(!pystat2dir(dt, &d))
		return nil;
	Py_BEGIN_ALLOW_THREADS
	ret = dirfwstat(fd, &d);
	Py_END_ALLOW_THREADS
	if(ret < 0)
		return plan9_error();
	Py_INCREF(Py_None);
	return Py_None;
}

static char plan9_remove__doc__[] =
"remove(path) -> None\n\
Remove a file.";

static PyObject *
plan9_remove(PyObject *self, PyObject *args)
{
	char *path;
	int res;

	if (!PyArg_ParseTuple(args, "s:remove", &path))
		return nil;
	Py_BEGIN_ALLOW_THREADS
	res = remove(path);
	Py_END_ALLOW_THREADS
	if (res < 0)
		return plan9_error();
	Py_INCREF(Py_None);
	return Py_None;
}

/* Process operations */
static char plan9__exits__doc__[] =
"_exits(status_string)\n\
Exit to the system with specified status, without normal exit processing.";

static PyObject *
plan9__exits(PyObject *self, PyObject *args)
{
	char *sts;
	if (!PyArg_ParseTuple(args, "s:_exits", &sts))
		return NULL;
	_exits(sts);
	return NULL; /* Make gcc -Wall happy */
}


/*
 * We can't call this exec because exec is a reserved word in Python.
 */

static char plan9__exec__doc__[] =
"_exec(path, args)\n\
Execute an executable path with arguments, replacing current process.\n\
\n\
	path: path of executable file\n\
	args: tuple or list of strings";

static PyObject *
plan9__exec(PyObject *self, PyObject *args)
{
	char *path;
	PyObject *argv;
	char **argvlist;
	int i, argc;
	PyObject *(*getitem)(PyObject *, int);

	/* exec has two arguments: (path, argv), where
	   argv is a list or tuple of strings. */

	if (!PyArg_ParseTuple(args, "sO:exec", &path, &argv))
		return NULL;
	if (PyList_Check(argv)) {
		argc = PyList_Size(argv);
		getitem = PyList_GetItem;
	}
	else if (PyTuple_Check(argv)) {
		argc = PyTuple_Size(argv);
		getitem = PyTuple_GetItem;
	}
	else {
		PyErr_SetString(PyExc_TypeError, "argv must be tuple or list");
		return NULL;
	}

	if (argc == 0) {
		PyErr_SetString(PyExc_ValueError, "empty argument list");
		return NULL;
	}

	argvlist = PyMem_NEW(char *, argc+1);
	if (argvlist == NULL)
		return NULL;
	for (i = 0; i < argc; i++) {
		if (!PyArg_Parse((*getitem)(argv, i), "s", &argvlist[i])) {
			PyMem_DEL(argvlist);
			PyErr_SetString(PyExc_TypeError, 
					"all arguments must be strings");
			return NULL;
			
		}
	}
	argvlist[argc] = NULL;

	exec(path, argvlist);

	/* If we get here it's definitely an error */
	PyMem_DEL(argvlist);
	return plan9_error();
}

static char plan9_rfork__doc__[] =
"rfork(flags) -> pid\n\
Manipulate process resources.\n\
\n\
If new child is created, return 0 to child process and PID of child to parent process.";

static PyObject *
plan9_rfork(PyObject *self, PyObject *args)
{
	int flags, pid;
	if (!PyArg_ParseTuple(args, "i:fork", &flags))
		return nil;
	pid = rfork(flags);
	if (pid == -1)
		return plan9_error();
	if (pid == 0)
		PyOS_AfterFork();
	return PyInt_FromLong((long)pid);
}

static char plan9_await__doc__[] =
"await() -> string\n\
Wait for process to exit, return process statistics.";

static PyObject*
plan9_await(PyObject *self, PyObject *args)
{
	char buf[ERRMAX+4*12+10];
	int n;

	if (!PyArg_ParseTuple(args, ":await"))
		return nil;

	n = await(buf, sizeof buf);
	if(n < 0)
		return plan9_error();
	return PyString_FromString(buf);
}


/* Functions acting on file descriptors */

static char plan9_open__doc__[] =
"open(filename, omode) -> fd\n\
Open a file (for low level IO).";

static PyObject *
plan9_open(PyObject *self, PyObject *args)
{
	char *file;
	int omode;
	int fd;
	if (!PyArg_ParseTuple(args, "si", &file, &omode))
		return nil;

	Py_BEGIN_ALLOW_THREADS
	fd = open(file, omode);
	Py_END_ALLOW_THREADS
	if (fd < 0)
		return plan9_error();
	return PyInt_FromLong((long)fd);
}

static char plan9_create__doc__[] =
"create(filename, omode, perm) -> fd\n\
Create a file (for low level IO).";

static PyObject *
plan9_create(PyObject *self, PyObject *args)
{
	char *path;
	int fd, omode, perm;

	if (!PyArg_ParseTuple(args, "sii", &path, &omode, &perm))
		return nil;

	Py_BEGIN_ALLOW_THREADS
	fd = create(path, omode, perm);
	Py_END_ALLOW_THREADS
	if (fd < 0)
		return plan9_error();
	return PyInt_FromLong((long)fd);
}

static char plan9_close__doc__[] =
"close(fd) -> None\n\
Close a file descriptor (for low level IO).";

static PyObject *
plan9_close(PyObject *self, PyObject *args)
{
	int fd, res;
	if (!PyArg_ParseTuple(args, "i:close", &fd))
		return nil;
	Py_BEGIN_ALLOW_THREADS
	res = close(fd);
	Py_END_ALLOW_THREADS
	if (res < 0)
		return plan9_error();
	Py_INCREF(Py_None);
	return Py_None;
}

static char plan9_fauth__doc__[] =
"fauth(fd, name) -> int\n\
Acquire an authentication fd.";

static PyObject *
plan9_fauth(PyObject *self, PyObject *args)
{
	int fd, res;
	char *name;

	if (!PyArg_ParseTuple(args, "is:fauth", &fd, &name))
		return nil;
	Py_BEGIN_ALLOW_THREADS
	res = fauth(fd, name);
	Py_END_ALLOW_THREADS
	if (res < 0)
		return plan9_error();
	return PyInt_FromLong(res);
}

static char plan9_fd2path__doc__[] =
"fd2path(fd) -> path\n\
Return the pathname used to open fd.";

static PyObject *
plan9_fd2path(PyObject *self, PyObject *args)
{
	int fd, res;
	char buf[1024];

	if (!PyArg_ParseTuple(args, "i:fd2path", &fd))
		return nil;
	Py_BEGIN_ALLOW_THREADS
	res = fd2path(fd, buf, sizeof buf);
	Py_END_ALLOW_THREADS
	if (res < 0)
		return plan9_error();
	return PyString_FromString(buf);
}

static char plan9_fversion__doc__[] =
"fversion(fd, bufsize, version-string) -> int\n\
Execute a 9P version transaction on fd.";

static PyObject *
plan9_fversion(PyObject *self, PyObject *args)
{
	int fd, msize, res;
	char buf[256];
	char *version;

	if (!PyArg_ParseTuple(args, "iis:fversion", &fd, &msize, &version))
		return nil;
	utfecpy(buf, buf+sizeof buf, version);
	Py_BEGIN_ALLOW_THREADS
	res = fversion(fd, msize, buf, sizeof buf);
	Py_END_ALLOW_THREADS
	if (res < 0)
		return plan9_error();
	return PyString_FromStringAndSize(buf, res);
}

static char plan9_dup__doc__[] =
"dup(old, new) -> new\n\
Duplicate file descriptor; system chooses if new==-1.";

static PyObject *
plan9_dup(PyObject *self, PyObject *args)
{
	int old, new, ret;
	if (!PyArg_ParseTuple(args, "ii:dup", &old, &new))
		return NULL;
	Py_BEGIN_ALLOW_THREADS
	new = dup(old, new);
	Py_END_ALLOW_THREADS
	if (new < 0)
		return plan9_error();
	return PyInt_FromLong((long)new);
}

static char plan9_seek__doc__[] =
"seek(fd, pos, how) -> newpos\n\
Set the current position of a file descriptor.";

static PyObject *
plan9_seek(PyObject *self, PyObject *args)
{
	int fd, how;
	vlong off;
	PyObject *o;

	if (!PyArg_ParseTuple(args, "iOi:seek", &fd, &o, &how)
	|| !getvlong(o, &off))
		return nil;
	
	Py_BEGIN_ALLOW_THREADS
	off = seek(fd, off, how);
	Py_END_ALLOW_THREADS
	if (off < 0)
		return plan9_error();
	return PyLong_FromLongLong(off);
}


static char plan9_pread__doc__[] =
"pread(fd, count, offset) -> string\n\
Read from a file descriptor.";

static PyObject *
plan9_pread(PyObject *self, PyObject *args)
{
	int fd, size, n;
	vlong off;

	PyObject *buffer;
	if (!PyArg_ParseTuple(args, "iiL:read", &fd, &size, &off))
		return nil;
	buffer = PyString_FromStringAndSize(nil, size);
	if (buffer == nil)
		return nil;
	Py_BEGIN_ALLOW_THREADS
	n = pread(fd, PyString_AsString(buffer), size, off);
	Py_END_ALLOW_THREADS
	if (n < 0) {
		Py_DECREF(buffer);
		return plan9_error();
	}
	if (n != size)
		_PyString_Resize(&buffer, n);
	return buffer;
}

static char plan9_pwrite__doc__[] =
"pwrite(fd, string, offset) -> byteswritten\n\
Write a string to a file descriptor.";

static PyObject *
plan9_pwrite(PyObject *self, PyObject *args)
{
	int fd, size;
	vlong off;
	char *buffer;

	if (!PyArg_ParseTuple(args, "is#L:write", &fd, &buffer, &size, &off))
		return nil;
	Py_BEGIN_ALLOW_THREADS
	size = pwrite(fd, buffer, size, off);
	Py_END_ALLOW_THREADS
	if (size < 0)
		return plan9_error();
	return PyInt_FromLong((long)size);
}

static char plan9_fdopen__doc__[] =
"fdopen(fd, [, mode='r' [, bufsize]]) -> file_object\n\
Return an open file object connected to a file descriptor.";

static PyObject *
plan9_fdopen(PyObject *self, PyObject *args)
{
	int fd;
	char *mode = "r";
	int bufsize = -1;
	FILE *fp;
	PyObject *f;
	if (!PyArg_ParseTuple(args, "i|si", &fd, &mode, &bufsize))
		return NULL;

	Py_BEGIN_ALLOW_THREADS
	fp = fdopen(fd, mode);
	Py_END_ALLOW_THREADS
	if (fp == NULL)
		return plan9_error();
	f = PyFile_FromFile(fp, "(fdopen)", mode, fclose);
	if (f != NULL)
		PyFile_SetBufSize(f, bufsize);
	return f;
}

static char plan9_pipe__doc__[] =
"pipe() -> (end0, end1)\n\
Create a pipe.";

static PyObject *
plan9_pipe(PyObject *self, PyObject *args)
{
	int fds[2];
	int res;
	if (!PyArg_ParseTuple(args, ":pipe"))
		return nil;
	Py_BEGIN_ALLOW_THREADS
	res = pipe(fds);
	Py_END_ALLOW_THREADS
	if (res != 0)
		return plan9_error();
	return Py_BuildValue("(ii)", fds[0], fds[1]);
}

static char plan9_putenv__doc__[] =
"putenv(key, value) -> None\n\
Change or add an environment variable.";

/* Save putenv() parameters as values here, so we can collect them when they
 * get re-set with another call for the same key. */
static PyObject *plan9_putenv_garbage;

static PyObject * 
plan9_putenv(PyObject *self, PyObject *args)
{
	char buf[256];
	char *key, *val;
	int fd;
	PyObject *newstr;

	if (!PyArg_ParseTuple(args, "ss:putenv", &key, &val))
		return nil;

	snprint(buf, sizeof buf, "#e/%s", key);
	if((fd = create(buf, OWRITE, 066)) < 0)
		return plan9_error();

	if(write(fd, val, strlen(val)) != strlen(val)){
		close(fd);
		return plan9_error();
	}
	close(fd);

	/* Install the first arg and newstr in plan9_putenv_garbage;
	 * this will cause previous value to be collected.  This has to
	 * happen after the real putenv() call because the old value
	 * was still accessible until then. */
	if (PyDict_SetItem(plan9_putenv_garbage,
			   PyTuple_GET_ITEM(args, 0), newstr))
		/* really not much we can do; just leak */
		PyErr_Clear();
	else
		Py_DECREF(newstr);

        return Py_None;
}

static char plan9_unmount__doc__[] =
"unmount(old, name) -> None\n\
Unmount name from old.";

static PyObject *
plan9_unmount(PyObject *self, PyObject *args)
{
	char *old, *name;
	int ret;

	if(!PyArg_ParseTuple(args, "ss:unmount", &old, &name))
		return nil;
	Py_BEGIN_ALLOW_THREADS
	ret = unmount(old, name[0] ? name : nil);
	Py_END_ALLOW_THREADS
	if(ret < 0)
		return plan9_error();
	return PyInt_FromLong(ret);
}

/* Copy from Python2.4 Moroo */
PyDoc_STRVAR(plan9_mkdir__doc__,
"mkdir(path [, mode=0777])\n\n\
Create a directory.");

static PyObject *
plan9_mkdir(PyObject *self, PyObject *args)
{
	int res;
	char *path = NULL;
	int mode = 0777;

	if (!PyArg_ParseTuple(args, "et|i:mkdir",
	                      Py_FileSystemDefaultEncoding, &path, &mode))
		return NULL;

	/* Following statements came from cmd/mkdir.c */
	if(access(path, AEXIST) == 0){
		PyMem_Free(path);
		return plan9_error();
	}
	Py_BEGIN_ALLOW_THREADS
	res = create(path, OREAD, DMDIR | mode);
	Py_END_ALLOW_THREADS
	PyMem_Free(path);
	if (res < 0)
		return plan9_error();
	close(res);
	Py_INCREF(Py_None);

	return Py_None;
}

static char plan9_abort__doc__[] = "\
abort() -> does not return!\n\
Abort the interpreter immediately.  This 'dumps core' or otherwise fails\n\
in the hardest way possible on the hosting operating system.";

static PyObject *
plan9_abort(PyObject *self, PyObject *args)
{
	if (!PyArg_ParseTuple(args, ":abort"))
		return nil;
	abort();
	/*NOTREACHED*/
	Py_FatalError("abort() called from Python code didn't abort!");
	return nil;
}

static char plan9_rendezvous__doc__[] = "\
rendezvous(tag, value) -> value\n\
plan 9 rendezvous call.";

static PyObject*
plan9_rendezvous(PyObject *self, PyObject *args)
{
	void *tag,  *val;

	if(!PyArg_ParseTuple(args, "ll:rendezvous", &tag, &val))
		return nil;
	Py_BEGIN_ALLOW_THREADS
	val = rendezvous((void *)tag, (void *)val);
	Py_END_ALLOW_THREADS
	if(val == (void *)(~0))
		return plan9_error();
	return PyInt_FromLong((ulong)val);
}

static char plan9_sleep__doc__[] = "\
sleep(millisec) -> None\n\
pause execution.";

static PyObject*
plan9_sleep(PyObject *self, PyObject *args)
{
	int n, ret;

	if(!PyArg_ParseTuple(args, "i:sleep", &n))
		return nil;
	Py_BEGIN_ALLOW_THREADS
	ret = sleep(n);
	Py_END_ALLOW_THREADS
	if(ret < 0)
		return plan9_error();
	Py_INCREF(Py_None);
	return Py_None;
}

static char*
qtoken(char *s, char *sep)
{
	int quoting;
	char *t;

	quoting = 0;
	t = s;	/* s is output string, t is input string */
	while(*t!='\0' && (quoting || utfrune(sep, *t)==nil)){
		if(*t != '\''){
			*s++ = *t++;
			continue;
		}
		/* *t is a quote */
		if(!quoting){
			quoting = 1;
			t++;
			continue;
		}
		/* quoting and we're on a quote */
		if(t[1] != '\''){
			/* end of quoted section; absorb closing quote */
			t++;
			quoting = 0;
			continue;
		}
		/* doubled quote; fold one quote into two */
		t++;
		*s++ = *t++;
	}
	if(*s != '\0'){
		*s = '\0';
		if(t == s)
			t++;
	}
	return t;
}

static char plan9_gettokens__doc__[] = "\
gettokens(s, sep) -> list of strings";

static PyObject*
plan9_gettokens(PyObject *self, PyObject *args)
{
	char *s, *os, *ms, *sep, **f;
	int nf, mf;
	PyObject *pd, *v;

	if(!PyArg_ParseTuple(args, "ss:gettokens", &s, &sep))
		return nil;
	if ((pd = PyList_New(0)) == nil)
		return nil;
	if ((s = strdup(s)) == nil){
		Py_DECREF(pd);
		return plan9_error();
	}
	ms = s;
	for(;;){
		while(*s!='\0' && utfrune(sep, *s)!=nil)
			s++;
		if(*s == '\0')
			break;
		os = s;
		s = qtoken(os, sep);
		v = PyString_FromString(os);
		if(v == nil){
			Py_DECREF(pd);
			free(s);
			return nil;
		}
		if(PyList_Append(pd, v) != 0){
			Py_DECREF(v);
			Py_DECREF(pd);
			free(s);
			return nil;
		}
	}
	free(ms);
	return pd;
}

static PyMethodDef plan9_methods[] = {
/* system calls */
	{"_exits",	plan9__exits, METH_VARARGS, plan9__exits__doc__},
	{"abort",	plan9_abort, METH_VARARGS, plan9_abort__doc__},
	{"access",	plan9_access, METH_VARARGS, plan9_access__doc__},
	{"await",	plan9_await, METH_VARARGS, plan9_await__doc__},
	{"bind",	plan9_bind, METH_VARARGS, plan9_bind__doc__},
	{"chdir",	plan9_chdir, METH_VARARGS, plan9_chdir__doc__},
	{"close",	plan9_close, METH_VARARGS, plan9_close__doc__},
	{"create",	plan9_create, METH_VARARGS, plan9_create__doc__},
	{"dup",		plan9_dup, METH_VARARGS, plan9_dup__doc__},
	{"_exec",	plan9__exec, METH_VARARGS, plan9__exec__doc__},
	{"fauth",	plan9_fauth, METH_VARARGS, plan9_fauth__doc__},
	{"fstat",	plan9_fstat, METH_VARARGS, plan9_fstat__doc__},
	{"fwstat",	plan9_fwstat, METH_VARARGS, plan9_fwstat__doc__},
	{"fversion",	plan9_fversion, METH_VARARGS, plan9_fversion__doc__},
	{"mount",	plan9_mount, METH_VARARGS, plan9_mount__doc__},
	{"unmount",	plan9_unmount, METH_VARARGS, plan9_unmount__doc__},
	{"open",	plan9_open, METH_VARARGS, plan9_open__doc__},
	{"fd2path",	plan9_fd2path, METH_VARARGS, plan9_fd2path__doc__},
	{"pipe",	plan9_pipe, METH_VARARGS, plan9_pipe__doc__},
	{"pread",	plan9_pread, METH_VARARGS, plan9_pread__doc__},
	{"pwrite",	plan9_pwrite, METH_VARARGS, plan9_pwrite__doc__},
	{"remove",	plan9_remove, METH_VARARGS, plan9_remove__doc__},
	{"rendezvous",	plan9_rendezvous, METH_VARARGS, plan9_rendezvous__doc__},
	{"rfork",	plan9_rfork, METH_VARARGS, plan9_rfork__doc__},
	{"seek",	plan9_seek, METH_VARARGS, plan9_seek__doc__},
	{"sleep",	plan9_sleep, METH_VARARGS, plan9_sleep__doc__},
	{"stat",	plan9_stat, METH_VARARGS, plan9_stat__doc__},
	{"wstat",	plan9_wstat, METH_VARARGS, plan9_wstat__doc__},

/* library calls */
	{"dirread",	plan9_dirread, METH_VARARGS, plan9_dirread__doc__},
	{"dirreadall", plan9_dirreadall, METH_VARARGS, plan9_dirreadall__doc__},
	{"gettokens", plan9_gettokens, METH_VARARGS, plan9_gettokens__doc__},

/* python additions */
	{"fdopen",	plan9_fdopen, METH_VARARGS, plan9_fdopen__doc__},

	{"mkdir",	plan9_mkdir, METH_VARARGS, plan9_mkdir__doc__},
//	{"rc",	plan9_rc, METH_VARARGS, plan9_rc__doc__},
	{nil,		nil}		 /* Sentinel */
};

/* This function copy from Python 2.4 Moroo */
static int
ins(PyObject *module, char *symbol, long value)
{
        return PyModule_AddIntConstant(module, symbol, value);
}

/* This function copy from Python 2.4 Moroo */
static int
all_ins(PyObject *d)
{
#ifdef F_OK
        if (ins(d, "F_OK", (long)F_OK)) return -1;
#endif
#ifdef R_OK
        if (ins(d, "R_OK", (long)R_OK)) return -1;
#endif
#ifdef W_OK
        if (ins(d, "W_OK", (long)W_OK)) return -1;
#endif
#ifdef X_OK
        if (ins(d, "X_OK", (long)X_OK)) return -1;
#endif
#ifdef NGROUPS_MAX
        if (ins(d, "NGROUPS_MAX", (long)NGROUPS_MAX)) return -1;
#endif
#ifdef TMP_MAX
        if (ins(d, "TMP_MAX", (long)TMP_MAX)) return -1;
#endif
#ifdef WCONTINUED
        if (ins(d, "WCONTINUED", (long)WCONTINUED)) return -1;
#endif
#ifdef WNOHANG
        if (ins(d, "WNOHANG", (long)WNOHANG)) return -1;
#endif
#ifdef WUNTRACED
        if (ins(d, "WUNTRACED", (long)WUNTRACED)) return -1;
#endif
#ifdef O_RDONLY
        if (ins(d, "O_RDONLY", (long)O_RDONLY)) return -1;
#endif
#ifdef O_WRONLY
        if (ins(d, "O_WRONLY", (long)O_WRONLY)) return -1;
#endif
#ifdef O_RDWR
        if (ins(d, "O_RDWR", (long)O_RDWR)) return -1;
#endif
#ifdef O_NDELAY
        if (ins(d, "O_NDELAY", (long)O_NDELAY)) return -1;
#endif
#ifdef O_NONBLOCK
        if (ins(d, "O_NONBLOCK", (long)O_NONBLOCK)) return -1;
#endif
#ifdef O_APPEND
        if (ins(d, "O_APPEND", (long)O_APPEND)) return -1;
#endif
#ifdef O_DSYNC
        if (ins(d, "O_DSYNC", (long)O_DSYNC)) return -1;
#endif
#ifdef O_RSYNC
        if (ins(d, "O_RSYNC", (long)O_RSYNC)) return -1;
#endif
#ifdef O_SYNC
        if (ins(d, "O_SYNC", (long)O_SYNC)) return -1;
#endif
#ifdef O_NOCTTY
        if (ins(d, "O_NOCTTY", (long)O_NOCTTY)) return -1;
#endif
#ifdef O_CREAT
        if (ins(d, "O_CREAT", (long)O_CREAT)) return -1;
#endif
#ifdef O_EXCL
        if (ins(d, "O_EXCL", (long)O_EXCL)) return -1;
#endif
#ifdef O_TRUNC
        if (ins(d, "O_TRUNC", (long)O_TRUNC)) return -1;
#endif
#ifdef O_BINARY
        if (ins(d, "O_BINARY", (long)O_BINARY)) return -1;
#endif
#ifdef O_TEXT
        if (ins(d, "O_TEXT", (long)O_TEXT)) return -1;
#endif
#ifdef O_LARGEFILE
        if (ins(d, "O_LARGEFILE", (long)O_LARGEFILE)) return -1;
#endif

/* MS Windows */
#ifdef O_NOINHERIT
	/* Don't inherit in child processes. */
        if (ins(d, "O_NOINHERIT", (long)O_NOINHERIT)) return -1;
#endif
#ifdef _O_SHORT_LIVED
	/* Optimize for short life (keep in memory). */
	/* MS forgot to define this one with a non-underscore form too. */
        if (ins(d, "O_SHORT_LIVED", (long)_O_SHORT_LIVED)) return -1;
#endif
#ifdef O_TEMPORARY
	/* Automatically delete when last handle is closed. */
        if (ins(d, "O_TEMPORARY", (long)O_TEMPORARY)) return -1;
#endif
#ifdef O_RANDOM
	/* Optimize for random access. */
        if (ins(d, "O_RANDOM", (long)O_RANDOM)) return -1;
#endif
#ifdef O_SEQUENTIAL
	/* Optimize for sequential access. */
        if (ins(d, "O_SEQUENTIAL", (long)O_SEQUENTIAL)) return -1;
#endif

/* GNU extensions. */
#ifdef O_DIRECT
        /* Direct disk access. */
        if (ins(d, "O_DIRECT", (long)O_DIRECT)) return -1;
#endif
#ifdef O_DIRECTORY
        /* Must be a directory.	 */
        if (ins(d, "O_DIRECTORY", (long)O_DIRECTORY)) return -1;
#endif
#ifdef O_NOFOLLOW
        /* Do not follow links.	 */
        if (ins(d, "O_NOFOLLOW", (long)O_NOFOLLOW)) return -1;
#endif

	/* These come from sysexits.h */
#ifdef EX_OK
	if (ins(d, "EX_OK", (long)EX_OK)) return -1;
#endif /* EX_OK */
#ifdef EX_USAGE
	if (ins(d, "EX_USAGE", (long)EX_USAGE)) return -1;
#endif /* EX_USAGE */
#ifdef EX_DATAERR
	if (ins(d, "EX_DATAERR", (long)EX_DATAERR)) return -1;
#endif /* EX_DATAERR */
#ifdef EX_NOINPUT
	if (ins(d, "EX_NOINPUT", (long)EX_NOINPUT)) return -1;
#endif /* EX_NOINPUT */
#ifdef EX_NOUSER
	if (ins(d, "EX_NOUSER", (long)EX_NOUSER)) return -1;
#endif /* EX_NOUSER */
#ifdef EX_NOHOST
	if (ins(d, "EX_NOHOST", (long)EX_NOHOST)) return -1;
#endif /* EX_NOHOST */
#ifdef EX_UNAVAILABLE
	if (ins(d, "EX_UNAVAILABLE", (long)EX_UNAVAILABLE)) return -1;
#endif /* EX_UNAVAILABLE */
#ifdef EX_SOFTWARE
	if (ins(d, "EX_SOFTWARE", (long)EX_SOFTWARE)) return -1;
#endif /* EX_SOFTWARE */
#ifdef EX_OSERR
	if (ins(d, "EX_OSERR", (long)EX_OSERR)) return -1;
#endif /* EX_OSERR */
#ifdef EX_OSFILE
	if (ins(d, "EX_OSFILE", (long)EX_OSFILE)) return -1;
#endif /* EX_OSFILE */
#ifdef EX_CANTCREAT
	if (ins(d, "EX_CANTCREAT", (long)EX_CANTCREAT)) return -1;
#endif /* EX_CANTCREAT */
#ifdef EX_IOERR
	if (ins(d, "EX_IOERR", (long)EX_IOERR)) return -1;
#endif /* EX_IOERR */
#ifdef EX_TEMPFAIL
	if (ins(d, "EX_TEMPFAIL", (long)EX_TEMPFAIL)) return -1;
#endif /* EX_TEMPFAIL */
#ifdef EX_PROTOCOL
	if (ins(d, "EX_PROTOCOL", (long)EX_PROTOCOL)) return -1;
#endif /* EX_PROTOCOL */
#ifdef EX_NOPERM
	if (ins(d, "EX_NOPERM", (long)EX_NOPERM)) return -1;
#endif /* EX_NOPERM */
#ifdef EX_CONFIG
	if (ins(d, "EX_CONFIG", (long)EX_CONFIG)) return -1;
#endif /* EX_CONFIG */
#ifdef EX_NOTFOUND
	if (ins(d, "EX_NOTFOUND", (long)EX_NOTFOUND)) return -1;
#endif /* EX_NOTFOUND */

#ifdef HAVE_SPAWNV
#if defined(PYOS_OS2) && defined(PYCC_GCC)
	if (ins(d, "P_WAIT", (long)P_WAIT)) return -1;
	if (ins(d, "P_NOWAIT", (long)P_NOWAIT)) return -1;
	if (ins(d, "P_OVERLAY", (long)P_OVERLAY)) return -1;
	if (ins(d, "P_DEBUG", (long)P_DEBUG)) return -1;
	if (ins(d, "P_SESSION", (long)P_SESSION)) return -1;
	if (ins(d, "P_DETACH", (long)P_DETACH)) return -1;
	if (ins(d, "P_PM", (long)P_PM)) return -1;
	if (ins(d, "P_DEFAULT", (long)P_DEFAULT)) return -1;
	if (ins(d, "P_MINIMIZE", (long)P_MINIMIZE)) return -1;
	if (ins(d, "P_MAXIMIZE", (long)P_MAXIMIZE)) return -1;
	if (ins(d, "P_FULLSCREEN", (long)P_FULLSCREEN)) return -1;
	if (ins(d, "P_WINDOWED", (long)P_WINDOWED)) return -1;
	if (ins(d, "P_FOREGROUND", (long)P_FOREGROUND)) return -1;
	if (ins(d, "P_BACKGROUND", (long)P_BACKGROUND)) return -1;
	if (ins(d, "P_NOCLOSE", (long)P_NOCLOSE)) return -1;
	if (ins(d, "P_NOSESSION", (long)P_NOSESSION)) return -1;
	if (ins(d, "P_QUOTE", (long)P_QUOTE)) return -1;
	if (ins(d, "P_TILDE", (long)P_TILDE)) return -1;
	if (ins(d, "P_UNRELATED", (long)P_UNRELATED)) return -1;
	if (ins(d, "P_DEBUGDESC", (long)P_DEBUGDESC)) return -1;
#else
        if (ins(d, "P_WAIT", (long)_P_WAIT)) return -1;
        if (ins(d, "P_NOWAIT", (long)_P_NOWAIT)) return -1;
        if (ins(d, "P_OVERLAY", (long)_OLD_P_OVERLAY)) return -1;
        if (ins(d, "P_NOWAITO", (long)_P_NOWAITO)) return -1;
        if (ins(d, "P_DETACH", (long)_P_DETACH)) return -1;
#endif
#endif

#if defined(PYOS_OS2)
        if (insertvalues(d)) return -1;
#endif
        return 0;
}

#define INITFUNC init_plan9
#define MODNAME "_plan9"

DL_EXPORT(void)
INITFUNC(void)
{
	PyObject *m, *v;

	m = Py_InitModule3(MODNAME,
			   plan9_methods,
			   plan9__doc__);

        if (all_ins(m))
                return;

	Py_INCREF(PyExc_OSError);
	PyModule_AddObject(m, "error", PyExc_OSError);

#ifdef HAVE_PUTENV
	if (plan9_putenv_garbage == NULL)
		plan9_putenv_garbage = PyDict_New();
#endif

}
