"""Pack and unpack 9P2000 messages."""

from __future__ import nested_scopes, generators
import sys, struct

class FcallError(Exception):
	def __init__(self, msg=None):
		if msg:
			self.msg = msg

IOHDRSZ = 24

Tversion, \
Rversion, \
Tauth, \
Rauth, \
Tattach, \
Rattach, \
Terror, \
Rerror, \
Tflush, \
Rflush, \
Twalk, \
Rwalk, \
Topen, \
Ropen, \
Tcreate, \
Rcreate, \
Tread, \
Rread, \
Twrite, \
Rwrite, \
Tclunk, \
Rclunk, \
Tremove, \
Rremove, \
Tstat, \
Rstat, \
Twstat, \
Rwstat = range(100, 128)

fcalls = {
	Tversion: ('LS', ['msize', 'version']),
	Rversion: ('LS', ['msize', 'version']),

	Tauth: ('LSS', ['afid', 'uname', 'aname']),
	Rauth: ('q', ['aqid']),

	Rerror: ('S', ['ename']),

	Tflush: ('H', ['oldtag']),
	Rflush: ('', []),

	Tattach: ('LLSS', ['fid', 'afid', 'uname', 'aname']),
	Rattach: ('Q', ['qid']),

	Twalk: ('FF#S', ['fid', 'newfid', 'wname']),
	Rwalk: ('#Q', ['qid']),

	Topen: ('LB', ['fid', 'mode']),
	Ropen: ('qL', ['qid', 'iounit']),

	Tcreate: ('LSLB', ['fid', 'name', 'perm', 'mode']),
	Rcreate: ('qL', ['qid', 'iounit']),

	Tread: ('LQL', ['fid', 'offset', 'count']),
	Rread: ('D', ['data']),

	Twrite: ('LQD', ['fid', 'offset', 'data']),
	Rwrite: ('L', ['count']),

	Tclunk: ('L', ['fid']),
	Rclunk: ('', []),

	Tremove: ('L', ['fid']),
	Rremove: ('', []),

	Tstat: ('L', ['fid']),
	Rstat: ('D', ['stat']),

	Twstat: ('LD', ['fid', 'stat']),
	Rwstat: ('', [])
}

fcallnames = {
	Tversion: 'Tversion',
	Rversion: 'Rversion',

	Tauth: 'Tauth',
	Rauth: 'Rauth',

	Rerror: 'Rerror',

	Tflush: 'Tflush',
	Rflush: 'Rflush',

	Tattach: 'Tattach',
	Rattach: 'Rattach',

	Twalk: 'Twalk',
	Rwalk: 'Rwalk',

	Topen: 'Topen',
	Ropen: 'Ropen',

	Tcreate: 'Tcreate',
	Rcreate: 'Rcreate',

	Tread: 'Tread',
	Rread: 'Rread',

	Twrite: 'Twrite',
	Rwrite: 'Rwrite',

	Tclunk: 'Tclunk',
	Rclunk: 'Rclunk',

	Tremove: 'Tremove',
	Rremove: 'Rremove',

	Tstat: 'Tstat',
	Rstat: 'Rstat',

	Twstat: 'Twstat',
	Rwstat: 'Rwstat',
}

def pack(fmt, vals):
	pkt = ''
	while fmt:
		c = fmt[0]
		fmt = fmt[1:]
		v = vals[0]
		vals = vals[1:]
		if c in 'BHLQ':
			pkt += struct.pack(c, v)
		elif c == 'S':
			pkt += pack('H', (len(v),))
			pkt += v
		elif c == 'q':
			pkt += pack('VLB', v.path, v.vers, v.type)
		elif c == 'D':
			pkt += pack('L', (len(v),))
			pkt += v
		elif c == '#':
			c = fmt[0]
			fmt = fmt[1:]
			pkt += pack(c*len(v), v)
		else:
			raise Exception('unknown format character \''+c+'\'')
	return pkt

def unpack(fmt, pkt):
	l = []
	while fmt:
		c = fmt[0]
		fmt = fmt[1:]
		if c == 'B':
			v, = struct.unpack('<B', pkt[:1])
			pkt = pkt[1:]
		elif c == 'H':
			v, = struct.unpack('<H', pkt[:2])
			pkt = pkt[2:]
		elif c == 'L':
			v, = struct.unpack('<L', pkt[:4])
			pkt = pkt[4:]
		elif c == 'Q':
			v, = struct.unpack('<Q', pkt[:8])
			pkt = pkt[8:]
		elif c == 'S':
			n, = struct.unpack('<H', pkt[:2])
			v = pkt[2:2+n]
			pkt = pkt[2+n:]
		elif c == 'q':
			q, pkt = unpack('VLB', pkt[:8+4+1])
			v = plan9.Qid().fromtuple(q)
		elif c == 'D':
			n, = struct.unpack('<L', pkt[:4])
			v = pkt[4:4+n]
			pkt = pkt[4+n:]
		elif c == '#':
			n, = struct.unpack('<H', pkt[:2])
			c = fmt[0]
			fmt = fmt[1:]
			v, pkt = unpack(c*n, pkt[2:])
		else:
			raise Exception('unknown format character \''+c+'\'')
		l.append(v)
	return l, pkt

def mkfcall(*args):
	f = Fcall()
	f.loadtuple(*args)
	return f

class Fcall:
	def __init__(self, type=None, args=None):
		if type==None:
			return
		if not fcalls.has_key(type):
			raise Exception('bad fcall type')
		self.type = type
		if args != None:
			self.fromtuple(args)
		return

	def loadtuple(self, *args):
		self.type = args[0]
		self.tag = args[1]
		names = fcalls[self.type][1]
		off = 2
		for n in names:
			self.__dict__[n] = args[off]
			off += 1
		return self

	def load(self, pkt):
		(l, self.type, self.tag), npkt = unpack('LBH', pkt)
		if l != len(pkt):
			raise Exception('packet length field is a lie')
		pkt = npkt
		if not fcalls.has_key(self.type):
			raise Exception('bad packet type %d' % (self.type,))
		fmt, names = fcalls[self.type]
		vals, pkt = unpack(fmt, pkt)
		if pkt:
			raise Exception('packet too large')
		for n,v in zip(names, vals):
			self.__dict__[n] = v
		return self
			
	def unload(self):
		hdr = pack('BH', (self.type, self.tag))
		fmt, names = fcalls[self.type]
		dat = pack(fmt, map(lambda x: self.__dict__[x], names))
		l = pack('L', (4+len(hdr)+len(dat),))
		return l+hdr+dat

	def __repr__(self):
		try:
			s = fcallnames[self.type]
			for n in fcalls[self.type][1]:
				s += ' '+str(self.__dict__[n])
			return s
		except Exception:
			return '<malformed or uninitialized Fcall>'

