"""OS routines for Plan 9 from Bell Labs

System calls are implemented by the _plan9 module.
Various wrappers are provided here.

"""

from _plan9 import *
from _plan9 import _exits, _exec
from plan9net import *

import string

MORDER	= 0x0003		# mask for bits defining order of mounting
MREPL	= 0x0000		# mount replaces object
MBEFORE	= 0x0001		# mount goes before others in union
MAFTER	= 0x0002		# mount goes after others in union
MCREATE	= 0x0004		# permit creation in mounted directory
MCACHE	= 0x0010		# cache some data
MMASK	= 0x0017		# all bits on

OREAD	= 0			# open for read
OWRITE	= 1			# write
ORDWR	= 2			# read and write
OEXEC	= 3			# execute, == read but check execute permission
OTRUNC	= 16			# or'ed in (except for exec), truncate file first
OCEXEC	= 32			# or'ed in, close on exec
ORCLOSE	= 64			# or'ed in, remove on close
OEXCL	= 0x1000		# or'ed in, exclusive use (create only)

AEXIST	= 0			# accessible: exists 
AEXEC	= 1			# execute access
AWRITE	= 2			# write access
AREAD	= 4			# read access

QTDIR	= 0x80		# type bit for directories
QTAPPEND = 0x40		# type bit for append only files
QTEXCL	= 0x20		# type bit for exclusive use files
QTMOUNT = 0x10		# type bit for mounted channel
QTAUTH	= 0x08		# type bit for authentication file
QTFILE	= 0x00		# plan file

DMDIR	= 0x80000000	# mode bit for directories
DMAPPEND = 0x40000000 # mode bit for append only files
DMEXCL	= 0x20000000	# mode bit for exclusive use files
DMMOUNT = 0x10000000 # mode bit for mounted channel
DMAUTH	= 0x08000000	# mode bit for authentication file
DMREAD	= 0x4		# mode bit for read permission
DMWRITE	= 0x2		# mode bit for write permission
DMEXEC	= 0x1		# mode bit for execute permission

RFNAMEG	= 1<<0		# new name space group
RFENVG	= 1<<1		# new environment group
RFFDG	= 1<<2		# new file descriptor group
RFNOTEG	= 1<<3		# new note group
RFPROC	= 1<<4		# new process
RFMEM	= 1<<5		# share memory
RFNOWAIT = 1<<6		# don't generate wait message for new process
RFCNAMEG = 1<<10		# clear name space
RFCENVG	= 1<<11		# clear environment
RFCFDG	= 1<<12		# clear file descriptors
RFREND	= 1<<13		# new rendezvous group
RFNOMNT = 1<<14		# don't allow child to mount

class Qid:
	def __init__(self):
		self.path = 0L
		self.vers = 0
		self.type = 0

	def fromtuple(self, x):
		self.path = x[0]
		self.vers = x[1]
		self.type = x[2]
		return self

	def totuple(self):
		return (self.path, self.vers, self.type)

	def __repr__(self):
		return '<Qid: 0x%lx, %d, 0x%x>' % (self.path, self.vers, self.type)

class Dir:
	def __init__(self):
		self.type = -1
		self.dev = '\x80'
		self.qid = Qid().fromtuple((-1, -1, 0xFF))
		self.mode = -1
		self.atime = -1
		self.mtime = -1
		self.length = -1
		self.name = ''
		self.uid = ''
		self.gid = ''
		self.muid = ''

	def fromtuple(self, x):
		d = Dir()
		d.name = x[0]
		d.qid = Qid().fromtuple(x[1])
		d.dev = x[2][0]
		d.type = x[2][1]
		d.mode = x[3]
		d.atime = x[4]
		d.mtime = x[5]
		d.length = x[6]
		d.uid = x[7]
		d.gid = x[8]
		d.muid = x[9]
		return d

	def totuple(self):
		return (self.name, self.qid.totuple(), (self.dev, self.type), self.mode, self.atime, \
			self.mtime, self.length, self.uid, self.gid, self.muid)

	def __repr__(self):
		return '<Dir: %s %s #%s%d %lo %ld %ld %ld %s %s %s>' % \
			(self.name, str(self.qid), self.dev, self.type, self.mode, self.atime, \
			self.mtime, self.length, self.uid, self.gid, self.muid)

class Waitmsg:
	def __init__(self):
		self.pid = -1
		self.utime = 0
		self.stime = 0
		self.rtime = 0
		self.msg = ''

	def fromtuple(self, x):
		self.pid = x[0]
		self.utime = x[1]
		self.stime = x[2]
		self.rtime = x[3]
		self.msg = x[4]
		return self

	def totuple(self):
		return (self.pid, self.utime, self.stime, self.rtime, self.msg)

	def __repr__(self):
		return '<Waitmsg: %d %ld %ld %ld %s>' % self.totuple()

# Moroo
class Statclass:
	def __init__(self,x):
		"""s is stat"""
		self.statdata = (x[3], x[1][1], x[2][0], 1, x[7], x[8], x[6], x[4], x[5], 0)
		self.st_mode = self.statdata[0]
		self.st_ino = self.statdata[1]
		self.st_dev = self.statdata[2]
		self.st_nlink = self.statdata[3]
		self.st_uid = self.statdata[4]
		self.st_gid = self.statdata[5]
		self.st_size = self.statdata[6]
		self.st_atime = self.statdata[7]
		self.st_mtime = self.statdata[8]
		self.st_ctime = self.statdata[9]

	def __len__(self):
		return len(self.statdata)

	def __getitem__(self,key):
		return self.statdata[key]

	def __cmp__(self,other):
		return cmp(self.statdata,other.statdata)
		
def dirstat(path):
	return Dir().fromtuple(stat(path))

def dirwstat(path, dir):
	return wstat(path, dir.totuple())

def dirfstat(fd):
	return Dir().fromtuple(fstat(fd))

def dirfwstat(fd, dir):
	return fwstat(fd, dir.totuple())

def fork():
	return rfork(RFFDG|RFREND|RFPROC)

# getcwd defined as getwd in os.py

# getlogin defined as getuser in os.py

def getnoteid():
	return string.atoi(readfile("/proc/"+`getpid()`+"/noteid"))

# getpgrp defined as getnoteid in os.py

def getpid():
	return string.atoi(readfile("/dev/pid"))

def getppid():
	return string.atoi(readfile("/dev/ppid"))

# getuid defined as getuser in os.py

def getuser():
	return readfile("/dev/user")

def getwd():
	try:
		fd = -1
		fd = open(".", OREAD)
		s = fd2path(fd)
		close(fd)
		return s
	except:
		close(fd)
		raise

def isatty(fd):
	s = fd2path(fd);
	return len(s)>=9 and s[-9:]=='/dev/cons'

# there are file descriptor leaks in all the popen routines
# error cases.  also, does f=fdopen(fd); f.close() close fd?

def popen(cmd, mode='r', bufsize=-1):
	if mode != 'r' and mode != 'w':
		raise ValueError('bad mode '+mode);
	t = pipe()
	pid = rfork(RFFDG|RFPROC|RFNOWAIT)
	if pid == 0:
		try:
			close(t[1])
			if mode == 'r':
				dup(t[0], 1)
			else:
				dup(t[0], 0)
			_exec('/bin/rc', ['rc', '-c', cmd])
		except Exception, msg:
			_exits(msg)
	else:
		close(t[0])
		return fdopen(t[0], mode, bufsize)
	
def popen2(cmd, mode='t', bufsize=-1):
	t = pipe()
	u = pipe()
	pid = rfork(RFFDG|RFPROC|RFNOWAIT)
	if pid == 0:
		try:
			close(t[1])
			close(u[1])
			dup(t[0], 0)
			dup(u[0], 1)
			_exec('/bin/rc', ['rc', '-c', cmd])
		except Exception, msg:
			_exits(msg)
	else:
		close(t[0])
		close(u[0])
		return fdopen(t[1], 'w', bufsize), fdopen(u[1], 'r', bufsize)

def popen3(cmd, mode='t', bufsize=-1):
	t = pipe()
	u = pipe()
	v = pipe()
	pid = rfork(RFFDG|RFPROC|RFNOWAIT)
	if pid == 0:
		try:
			close(t[1])
			close(u[1])
			close(v[1])
			dup(t[0], 0)
			dup(u[0], 1)
			dup(v[0], 2)
			_exec('/bin/rc', ['rc', '-c', cmd])
		except Exception, msg:
			_exits(msg)
	else:
		close(t[0])
		close(u[0])
		close(v[0])
		return fdopen(t[1], 'w', bufsize), fdopen(u[1], 'r', bufsize), fdopen(v[1], 'r', bufsize)

def popen4(cmd, mode='t', bufsize=-1):
	t = pipe()
	u = pipe()
	pid = rfork(RFFDG|RFPROC|RFNOWAIT)
	if pid == 0:
		try:
			close(t[1])
			close(u[1])
			dup(t[0], 0)
			dup(u[0], 1)
			dup(u[0], 2)
			_exec('/bin/rc', ['rc', '-c', cmd])
		except Exception, msg:
			_exits(msg)
	else:
		close(t[0])
		close(u[0])
		close(v[0])
		return fdopen(t[1], 'w', bufsize), fdopen(u[1], 'r', bufsize)	

def read(fd, len):
	return pread(fd, len, -1L)

def read9pmsg(fd):
	p = read(fd, 4)
	if len(p) != 4:
		raise Exception('short read')
	a = map(lambda x: ord(x), p)
	n = a[0] | (a[1]<<8) | (a[2]<<16) | (a[3]<<24)
	q = readn(fd, n-4)
	if len(q) != n-4:
		raise Exception('short read')
	return p+q

def readfile(path):
	fd = open(path, OREAD)
	tot = ''
	try:
		s = read(fd, 1024)
		while s != '':
			tot += s
			s = read(fd, 1024)
	finally:
		close(fd)
	return tot

def readn(fd, m):
	tot = ''
	while 1:
		if len(tot) >= m:
			break
		s = read(fd, m)
		tot += s
		if s == '':
			break
	return tot

def _splitpath(p):
	t = string.rfind(p, "/")+1
	if t == 0:
		return (".", p)
	return (p[0:t], p[t:])

def rename(old, new):
	(odir, oname) = _splitpath(old)
	(ndir, nname) = _splitpath(new)
	if odir != ndir:
		raise ValueError, "rename from dir "+odir+" to "+ndir
	d=Dir()
	d.name = nname
	return wstat(old, d.totuple())

def sysname():
	return readfile("/env/sysname")

def time():
	return string.atoi(tokenize(readfile("/dev/time"))[0])

def tokenize(s):
	return gettokens(s, " \t\r\n")

def write(fd, s):
	return pwrite(fd, s, -1L)

def wait():
	return Waitmsg().fromtuple(tokenize(await()))

def waitpid():
	return string.atoi(tokenize(await())[0])
	
# Implementations for the Posix module to import

def osstat(p):
	x = stat(p)
	#return (x[3], x[1][1], x[2][0], 1, x[7], x[8], x[6], x[4], x[5], 0)
	return Statclass(x)

def osdup(fd):
	return dup(fd, -1)

def listdir(p):
	l = dirreadall(p)
	return map(lambda x:x[0], l)

def uname():
	try:
		node=readfile("/dev/sysname")
	except:
		node="gnot"
	try:
		release="9P"+tokenize(readfile("/dev/osversion"))[0]
	except:
		release="9P1"
	try:
		version=readfile("/env/terminal")
	except:
		version=""
	try:
		version+=readfile("/env/bootfile")
	except:
		pass
	try:
		machine=readfile("/env/cputype")
	except:
		machine="unknown"
	return ("Plan 9", node, release, version, machine)

from UserDict import UserDict

class _Env(UserDict):
	def __init__(self, dict=None):
		UserDict.__init__(self, dict)
		self.coherence()
	def __setitem__(self, key, item):
		key = str(key)
		writefile('#e/'+key, _e2s(item))
		UserDict.__setitem__(self, key, item)
	def coherence(self):
		l = listdir('#e')
		for k in l:
			try:
				s = readfile('#e/'+k)
				if s=='' or s[len(s)-1] != '\x00':
					s += '\x00'
				v = s.split('\x00')
				v = v[:len(v)-1]
				if len(v)==1:
					v = v[0]
				self.data[k] = v
			except:
				pass

environ = _Env()